@@ -28,8 +28,8 @@
 #define PK_KEY_SIZE	16
 #define PK_CIPHER_BLOCK	16
 /* Auxiliary constants */
-#define RSA_DATA_SIZE	2000
-#define KEY_DATA_SIZE	1000
+#define RSA_DATA_SIZE	(512+16+512+256+256+256) /* n,e,d,p,q,u */
+#define KEY_DATA_SIZE	512
 #define RSA_FIELDS	6
 #define RSA_FIELDS_PUB	2
 #define KEY_TEST_SIZE	8
@@ -58,5 +58,5 @@ void read_password(char *pass_buf, int vrfy);
 void encrypt_init(unsigned char *, unsigned char *, char *);
 void get_random_salt(unsigned char *salt, size_t size);
 
-#define KEY_FILE	""
+#define SUSPEND_KEY_FILE	"/etc/suspend.key"
 #endif
