@@ -9,8 +9,7 @@
  *
  */
 
-#define _LARGEFILE64_SOURCE
-
+#include "config.h"
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/ioctl.h>
@@ -24,6 +23,7 @@
 #include <linux/kd.h>
 #include <linux/tiocl.h>
 #include <syscall.h>
+#include <libgen.h>
 #include <fcntl.h>
 #include <unistd.h>
 #include <stdio.h>
@@ -32,22 +32,25 @@
 #include <errno.h>
 #include <signal.h>
 #include <termios.h>
-#include <getopt.h>
 #ifdef CONFIG_COMPRESS
-#include <lzf.h>
-#else
-#define lzf_compress(a, b, c, d)	0
+#include <lzo/lzo1x.h>
 #endif
 
 #include "swsusp.h"
-#include "config.h"
+#include "config_parser.h"
 #include "md5.h"
 #include "splash.h"
 #include "vt.h"
+#include "loglevel.h"
 #ifdef CONFIG_BOTH
 #include "s2ram.h"
 #endif
 
+#define suspend_error(msg, args...) \
+do { \
+	fprintf(stderr, "%s: " msg " Reason: %m\n", my_name, ## args); \
+} while (0)
+
 static char snapshot_dev_name[MAX_STR_LEN] = SNAPSHOT_DEVICE;
 static char resume_dev_name[MAX_STR_LEN] = RESUME_DEVICE;
 static loff_t resume_offset;
@@ -55,18 +58,19 @@ static unsigned long pref_image_size = IMAGE_SIZE;
 static int suspend_loglevel = SUSPEND_LOGLEVEL;
 static char compute_checksum;
 #ifdef CONFIG_COMPRESS
-static char compress;
+static char do_compress;
+static long long compr_diff;
 #else
-#define compress 0
+#define do_compress 0
+#define compr_diff 0
 #endif
-static unsigned long compr_diff;
 #ifdef CONFIG_ENCRYPT
-static char encrypt;
+static char do_encrypt;
 static char use_RSA;
-static char key_name[MAX_STR_LEN] = KEY_FILE;
+static char key_name[MAX_STR_LEN] = SUSPEND_KEY_FILE;
 static char password[PASS_SIZE];
 #else
-#define encrypt 0
+#define do_encrypt 0
 #define key_name NULL
 #endif
 #ifdef CONFIG_BOTH
@@ -74,16 +78,24 @@ static char s2ram;
 #endif
 static char early_writeout;
 static char splash_param;
+#ifdef CONFIG_FBSPLASH
+char fbsplash_theme[MAX_STR_LEN] = "";
+#endif
 #define SHUTDOWN_LEN	16
-static char shutdown_method[SHUTDOWN_LEN] = "platform";
-static int use_platform_suspend;
+static char shutdown_method_value[SHUTDOWN_LEN] = "";
+static enum {
+	SHUTDOWN_METHOD_SHUTDOWN,
+	SHUTDOWN_METHOD_PLATFORM,
+	SHUTDOWN_METHOD_REBOOT
+} shutdown_method = SHUTDOWN_METHOD_PLATFORM;
 
 static int suspend_swappiness = SUSPEND_SWAPPINESS;
 static struct splash splash;
 static struct vt_mode orig_vtm;
 static int vfd;
+static char *my_name;
 
-static struct config_par parameters[PARAM_NO] = {
+static struct config_par parameters[] = {
 	{
 		.name = "snapshot device",
 		.fmt = "%s",
@@ -125,14 +137,14 @@ static struct config_par parameters[PARAM_NO] = {
 	{
 		.name = "compress",
 		.fmt = "%c",
-		.ptr = &compress,
+		.ptr = &do_compress,
 	},
 #endif
 #ifdef CONFIG_ENCRYPT
 	{
 		.name = "encrypt",
 		.fmt = "%c",
-		.ptr = &encrypt,
+		.ptr = &do_encrypt,
 	},
 	{
 		.name = "RSA key file",
@@ -154,9 +166,23 @@ static struct config_par parameters[PARAM_NO] = {
 	{
 		.name = "shutdown method",
 		.fmt = "%s",
-		.ptr = shutdown_method,
+		.ptr = shutdown_method_value,
 		.len = SHUTDOWN_LEN,
 	},
+#ifdef CONFIG_FBSPLASH
+	{
+		.name = "fbsplash theme",
+		.fmt = "%s",
+		.ptr = fbsplash_theme,
+		.len = MAX_STR_LEN,
+	},
+#endif
+	{
+		.name = NULL,
+		.fmt = NULL,
+		.ptr = NULL,
+		.len = 0,
+	}
 };
 
 static unsigned int page_size;
@@ -178,7 +204,7 @@ static inline loff_t check_free_swap(int dev)
 	if (!error)
 		return free_swap;
 	else
-		printf("Error: errno = %d\n", errno);
+		suspend_error("check_free_swap failed.");
 	return 0;
 }
 
@@ -198,7 +224,7 @@ static inline int free_swap_pages(int dev)
 	return ioctl(dev, SNAPSHOT_FREE_SWAP_PAGES, 0);
 }
 
-static inline int set_swap_file(int dev, dev_t blkdev, loff_t offset)
+static inline int set_swap_file(int dev, u_int32_t blkdev, loff_t offset)
 {
 	struct resume_swap_area swap;
 	int error;
@@ -226,7 +252,7 @@ static int write_area(int fd, void *buf, loff_t offset, unsigned int size)
 	ssize_t cnt = 0;
 
 	if (offset) {
-		if (lseek64(fd, offset, SEEK_SET) == offset)
+		if (lseek(fd, offset, SEEK_SET) == offset)
 			cnt = write(fd, buf, size);
 		if (cnt == size)
 			res = 0;
@@ -253,6 +279,9 @@ struct swap_map_handle {
 	unsigned int k;
 	int dev, fd;
 	struct md5_ctx ctx;
+#ifdef CONFIG_COMPRESS
+	void *lzo_work_buffer;
+#endif
 #ifdef CONFIG_ENCRYPT
 	unsigned char *encrypt_buffer;
 #endif
@@ -263,17 +292,30 @@ init_swap_writer(struct swap_map_handle *handle, int dev, int fd, void *buf)
 {
 	if (!buf)
 		return -EINVAL;
+
 	handle->areas = buf;
+	buf += page_size;
+
+	handle->page_buffer = buf;
+	buf += page_size;
+
+	handle->write_buffer = buf;
+	buf += buffer_size;
+
+#ifdef CONFIG_COMPRESS
+	handle->lzo_work_buffer = buf;
+	buf += LZO1X_1_MEM_COMPRESS;
+#endif
+
+#ifdef CONFIG_ENCRYPT
+	handle->encrypt_buffer = buf;
+#endif
+
+	memset(handle->areas, 0, page_size);
 	handle->areas_per_page = (page_size - sizeof(loff_t)) /
 			sizeof(struct swap_area);
 	handle->next_swap = (loff_t *)(handle->areas + handle->areas_per_page);
-	handle->page_buffer = (char *)buf + page_size;
-	handle->write_buffer = handle->page_buffer + page_size;
-#ifdef CONFIG_ENCRYPT
-	handle->encrypt_buffer = (unsigned char *)(handle->write_buffer +
-				buffer_size);
-#endif
-	memset(handle->areas, 0, page_size);
+
 	handle->cur_swap = get_swap_page(dev);
 	if (!handle->cur_swap)
 		return -ENOSPC;
@@ -283,10 +325,13 @@ init_swap_writer(struct swap_map_handle *handle, int dev, int fd, void *buf)
 	handle->cur_area.size = 0;
 	handle->cur_alloc = page_size;
 	handle->k = 0;
+
 	handle->dev = dev;
 	handle->fd = fd;
+
 	if (compute_checksum)
 		md5_init_ctx(&handle->ctx);
+
 	return 0;
 }
 
@@ -296,22 +341,22 @@ static int prepare(struct swap_map_handle *handle, int disp)
 	void *buf = handle->page_buffer;
 
 	block = (struct buf_block *)(handle->write_buffer + disp);
-	if (compress) {
-		unsigned short cnt;
+#ifdef CONFIG_COMPRESS
+	if (do_compress) {
+		lzo_uint cnt;
 
-		cnt = lzf_compress(buf, page_size,
-				block->data, page_size - sizeof(short));
-		if (!cnt) {
-			memcpy(block->data, buf, page_size);
-			cnt = page_size;
-		} else {
-			compr_diff += page_size - cnt;
-		}
-		compr_diff -= sizeof(short);
+		lzo1x_1_compress(buf, page_size, (lzo_bytep)block->data, &cnt,
+					handle->lzo_work_buffer);
 		block->size = cnt;
-		cnt += sizeof(short);
-		return cnt;
+		/* Sanity check, should not trigger */
+		if (cnt + sizeof(short) > max_block_size)
+			printf("WARNING!"
+				" Data size too large after compression!\n");
+		/* We have added a sizeof(short) field to the data */
+		compr_diff += page_size - cnt - sizeof(short);
+		return cnt + sizeof(short);
 	}
+#endif
 	memcpy(block, buf, page_size);
 	return page_size;
 }
@@ -340,7 +385,7 @@ static int flush_buffer(struct swap_map_handle *handle)
 	int error =  0;
 
 #ifdef CONFIG_ENCRYPT
-	if (encrypt) {
+	if (do_encrypt) {
 		error = gcry_cipher_encrypt(cipher_handle,
 			handle->encrypt_buffer, handle->cur_area.size,
 			src, handle->cur_area.size);
@@ -446,30 +491,22 @@ static int save_image(struct swap_map_handle *handle,
 	unsigned int m, writeout_rate;
 	int ret, abort_possible;
 	struct termios newtrm, savedtrm;
-	char c = 0;
 	int error = 0;
+	char message[SPLASH_GENERIC_MESSAGE_SIZE];
 
 	/* Switch the state of the terminal so that we can read the keyboard
 	 * without blocking and with no echo.
 	 *
 	 * stdin must be attached to the terminal now.
 	 */
-	abort_possible = !tcgetattr(0, &savedtrm);
-	if (abort_possible) {
-		newtrm = savedtrm;
-		newtrm.c_cc[VMIN] = 0;
-		newtrm.c_cc[VTIME] = 1;
-		newtrm.c_iflag = IGNBRK | IGNPAR | ICRNL | IMAXBEL;
-		newtrm.c_lflag = 0;
-		abort_possible = !tcsetattr(0, TCSANOW, &newtrm);
-	}
+	abort_possible = !splash.prepare_abort(&savedtrm, &newtrm);
+
+	sprintf(message, "Saving %u image data pages", nr_pages);
 	if (abort_possible)
-		printf("suspend: Saving %u image data pages "
-			"(press " ABORT_KEY_NAME " to abort) ...     ",
-			nr_pages);
-	else
-		printf("suspend: Saving image data pages (%u pages) ...     ",
-			nr_pages);
+		strcat(message, "(press " ABORT_KEY_NAME " to abort)");
+	strcat(message, "...");
+	printf("%s: %s     ", my_name, message);
+	splash.set_caption(message);
 
 	m = nr_pages / 100;
 	if (!m)
@@ -480,40 +517,50 @@ static int save_image(struct swap_map_handle *handle,
 	else
 		writeout_rate = nr_pages;
 
-	nr_pages = 0;
-	do {
+	for (nr_pages = 0; ; nr_pages++) {
 		ret = read(handle->dev, handle->page_buffer, page_size);
-		if (ret > 0) {
-			error = swap_write_page(handle);
-			if (error)
-				break;
+		if (ret <= 0)
+			break;
 
-			if (!(nr_pages % m)) {
-				printf("\b\b\b\b%3d%%", nr_pages / m);
-				splash.progress(20 + (nr_pages / m) * 0.75);
-				if (abort_possible) {
-					ret = read(0, &c, 1);
-					if (ret > 0 && c == ABORT_KEY_CODE) {
+		error = swap_write_page(handle);
+		if (error)
+			break;
+
+		if (!(nr_pages % m)) {
+			printf("\b\b\b\b%3d%%", nr_pages / m);
+			splash.progress(20 + (nr_pages / m) * 0.75);
+
+			switch (splash.key_pressed()) {
+				case ABORT_KEY_CODE:
+					if (abort_possible) {
 						printf(" aborted!\n");
-						return -EINTR;
+						error = -EINTR;
+						goto Exit;
 					}
-					ret = 1;
-				}
+					break;
+				case REBOOT_KEY_CODE:
+					printf (" reboot enabled\b\b\b\b\b\b\b"
+						"\b\b\b\b\b\b\b\b");
+					splash.set_caption("Reboot enabled");
+					shutdown_method =
+							SHUTDOWN_METHOD_REBOOT;
+					break;
 			}
-			if (!(nr_pages % writeout_rate))
-				start_writeout(handle->fd);
-
-			nr_pages++;
 		}
-	} while (ret > 0);
+
+		if (!(nr_pages % writeout_rate))
+			start_writeout(handle->fd);
+	}
+
 	if (ret < 0)
 		error = -errno;
 
 	if (!error)
 		printf(" done (%u pages)\n", nr_pages);
 
+Exit:
 	if (abort_possible)
-		tcsetattr(0, TCSANOW, &savedtrm);
+		splash.restore_abort(&savedtrm);
 
 	return error;
 }
@@ -529,7 +576,7 @@ static int enough_swap(int dev, unsigned long size)
 {
 	loff_t free_swap = check_free_swap(dev);
 
-	printf("suspend: Free swap: %lu kilobytes\n",
+	printf("%s: Free swap: %lu kilobytes\n", my_name,
 		(unsigned long)free_swap / 1024);
 	return free_swap > size;
 }
@@ -576,107 +623,131 @@ int write_image(int snapshot_fd, int resume_fd)
 	int error;
  	struct timeval begin;
 
-	printf("suspend: System snapshot ready. Preparing to write\n");
+	printf("%s: System snapshot ready. Preparing to write\n", my_name);
 	start = get_swap_page(snapshot_fd);
 	if (!start)
 		return -ENOSPC;
 	error = read(snapshot_fd, header, page_size);
 	if (error < (int)page_size)
 		return error < 0 ? error : -EFAULT;
-	printf("suspend: Image size: %lu kilobytes\n", header->size / 1024);
-	if (!enough_swap(snapshot_fd, header->size) && !compress) {
-		fprintf(stderr, "suspend: Not enough free swap\n");
+	printf("%s: Image size: %lu kilobytes\n", my_name, header->size / 1024);
+	if (!enough_swap(snapshot_fd, header->size) && !do_compress) {
+		fprintf(stderr, "%s: Not enough free swap\n", my_name);
 		return -ENOSPC;
 	}
 	error = init_swap_writer(&handle, snapshot_fd, resume_fd,
 			(char *)mem_pool + page_size);
-	if (!error) {
-		header->map_start = handle.cur_swap;
-		header->image_flags = 0;
-		max_block_size = page_size;
-		if (compute_checksum)
-			header->image_flags |= IMAGE_CHECKSUM;
-		if (compress) {
-			header->image_flags |= IMAGE_COMPRESSED;
-			max_block_size += sizeof(short);
-		}
+	if (error)
+		goto Exit;
+
+	header->map_start = handle.cur_swap;
+	header->image_flags = 0;
+	max_block_size = page_size;
+	if (compute_checksum)
+		header->image_flags |= IMAGE_CHECKSUM;
+
+	if (do_compress) {
+		header->image_flags |= IMAGE_COMPRESSED;
+		/*
+		 * The formula below follows from the worst-case expansion
+		 * calculation for LZO1.  sizeof(short) is due to the size field
+		 * in 'struct buf_block'.
+		 */
+		max_block_size += (page_size >> 4) + 67 + sizeof(short);
+	}
+
 #ifdef CONFIG_ENCRYPT
-		if (encrypt) {
-			if (use_RSA) {
-				error = gcry_cipher_setkey(cipher_handle,
-						key_data->key, KEY_SIZE);
-				if (error)
-					goto No_RSA;
+	if (!do_encrypt)
+		goto Save_image;
 
-				error = gcry_cipher_setiv(cipher_handle,
-						key_data->ivec, CIPHER_BLOCK);
+	if (use_RSA) {
+		error = gcry_cipher_setkey(cipher_handle, key_data->key,
+						KEY_SIZE);
+		if (error)
+			goto No_RSA;
 
-				if (error)
-					goto No_RSA;
+		error = gcry_cipher_setiv(cipher_handle, key_data->ivec,
+						CIPHER_BLOCK);
+		if (error)
+			goto No_RSA;
 
-				header->image_flags |= IMAGE_ENCRYPTED |
-							IMAGE_USE_RSA;
-				memcpy(&header->rsa, &key_data->rsa,
-						sizeof(struct RSA_data));
-				memcpy(&header->key, &key_data->encrypted_key,
+		header->image_flags |= IMAGE_ENCRYPTED | IMAGE_USE_RSA;
+		memcpy(&header->rsa, &key_data->rsa, sizeof(struct RSA_data));
+		memcpy(&header->key, &key_data->encrypted_key,
 						sizeof(struct encrypted_key));
-			} else {
-				int j;
+	} else {
+		int j;
 
 No_RSA:
-				encrypt_init(key_data->key, key_data->ivec,
-						password);
-				splash.progress(20);
-				get_random_salt(header->salt, CIPHER_BLOCK);
-				for (j = 0; j < CIPHER_BLOCK; j++)
-					key_data->ivec[j] ^= header->salt[j];
+		encrypt_init(key_data->key, key_data->ivec, password);
+		splash.progress(20);
+		get_random_salt(header->salt, CIPHER_BLOCK);
+		for (j = 0; j < CIPHER_BLOCK; j++)
+			key_data->ivec[j] ^= header->salt[j];
 
-				error = gcry_cipher_setkey(cipher_handle,
-						key_data->key, KEY_SIZE);
-				if (!error)
-					error = gcry_cipher_setiv(cipher_handle,
-						key_data->ivec, CIPHER_BLOCK);
+		error = gcry_cipher_setkey(cipher_handle, key_data->key,
+						KEY_SIZE);
+		if (!error)
+			error = gcry_cipher_setiv(cipher_handle, key_data->ivec,
+						CIPHER_BLOCK);
+		if (!error)
+			header->image_flags |= IMAGE_ENCRYPTED;
+	}
 
-				if (!error)
-					header->image_flags |= IMAGE_ENCRYPTED;
-			}
-			if (error)
-				fprintf(stderr,"suspend: libgcrypt error: %s\n",
-						gcry_strerror(error));
-		}
+	if (error) {
+		fprintf(stderr,"%s: libgcrypt error: %s\n", my_name,
+			gcry_strerror(error));
+		goto Exit;
+	}
+
+Save_image:
 #endif
-		if (!error) {
-			gettimeofday(&begin, NULL);
-			error = save_image(&handle, header->pages - 1);
-		}
-	}
-	if (!error) {
-		error = flush_swap_writer(&handle);
-		if (compute_checksum)
-			md5_finish_ctx(&handle.ctx, header->checksum);
-	}
-	fsync(resume_fd);
+	gettimeofday(&begin, NULL);
+
+	error = save_image(&handle, header->pages - 1);
+	if (error)
+		goto Exit;
+
+	/*
+	 * NOTICE: This needs to go after save_image(), because the user may
+	 * modify the behavior.
+	 */
+	if (shutdown_method == SHUTDOWN_METHOD_PLATFORM)
+		header->image_flags |= PLATFORM_SUSPEND;
+
+	error = flush_swap_writer(&handle);
 	if (!error) {
 		struct timeval end;
+
+		fsync(resume_fd);
+
+		if (compute_checksum)
+			md5_finish_ctx(&handle.ctx, header->checksum);
+
 		gettimeofday(&end, NULL);
 		timersub(&end, &begin, &end);
 		header->writeout_time = end.tv_usec / 1000000.0 + end.tv_sec;
-	}
-	if (!error)
+
 		error = write_area(resume_fd, header, start, page_size);
+	}
+
 	if (!error) {
-		if (compress) {
+		if (do_compress) {
 			double delta = header->size - compr_diff;
 
-			printf("suspend: Compression ratio %4.2lf\n",
+			printf("%s: Compression ratio %4.2lf\n", my_name,
 				delta / header->size);
 		}
 		printf( "S" );
 		error = mark_swap(resume_fd, start);
 	}
+
 	fsync(resume_fd);
+
 	if (!error)
 		printf( "|" );
+
+Exit:
 	printf("\n");
 	return error;
 }
@@ -728,19 +799,21 @@ static int reset_signature(int fd)
 
 static void suspend_shutdown(int snapshot_fd)
 {
-	if (!strcmp(shutdown_method, "reboot"))
+	splash.set_caption("Done.");
+
+	if (shutdown_method == SHUTDOWN_METHOD_REBOOT) {
 		reboot();
-	else if (use_platform_suspend) {
+	} else if (shutdown_method == SHUTDOWN_METHOD_PLATFORM) {
 		int ret = platform_enter(snapshot_fd);
 		if (ret < 0)
-			fprintf(stderr, "suspend: pm_ops->enter returned"
-				" error %d, calling power_off\n", ret);
+			suspend_error("pm_ops->enter failed, calling power_off.");
 	}
 	power_off();
 	/* Signature is on disk, it is very dangerous to continue now.
 	 * We'd do resume with stale caches on next boot. */
 	fprintf(stderr,"Powerdown failed. That's impossible.\n");
-	while(1);
+	while(1)
+		sleep (60);
 }
 
 int suspend_system(int snapshot_fd, int resume_fd)
@@ -748,6 +821,7 @@ int suspend_system(int snapshot_fd, int resume_fd)
 	loff_t avail_swap;
 	unsigned long image_size;
 	int attempts, in_suspend, error = 0;
+	char message[SPLASH_GENERIC_MESSAGE_SIZE];
 
 	avail_swap = check_free_swap(snapshot_fd);
 	if (avail_swap > pref_image_size)
@@ -755,7 +829,7 @@ int suspend_system(int snapshot_fd, int resume_fd)
 	else
 		image_size = avail_swap;
 	if (!avail_swap) {
-		fprintf(stderr, "suspend: No swap space for suspend\n");
+		fprintf(stderr, "%s: No swap space for suspend\n", my_name);
 		return ENOSPC;
 	}
 
@@ -770,66 +844,70 @@ int suspend_system(int snapshot_fd, int resume_fd)
 	if (error)
 		goto Unfreeze;
 
-	if (use_platform_suspend) {
+	if (shutdown_method == SHUTDOWN_METHOD_PLATFORM) {
 		int ret = platform_prepare(snapshot_fd);
-		if (ret < 0)
-			fprintf(stderr, "suspend: pm_ops->prepare returned "
-				"error %d\n", ret);
+		if (ret < 0) {
+			suspend_error("pm_ops->prepare failed, using "
+				"'shutdown mode = shutdown'.");
+			shutdown_method = SHUTDOWN_METHOD_SHUTDOWN;
+		}
 	}
 
-	printf("suspend: Snapshotting system\n");
+	sprintf(message, "Snapshotting system");
+	printf("%s: %s\n", my_name, message);
+	splash.set_caption(message);
 	attempts = 2;
 	do {
 		if (set_image_size(snapshot_fd, image_size)) {
 			error = errno;
 			break;
 		}
-		if (!atomic_snapshot(snapshot_fd, &in_suspend)) {
-			if (!in_suspend) {
-				free_snapshot(snapshot_fd);
-				if (use_platform_suspend)
-					platform_finish(snapshot_fd);
-				break;
-			}
-			error = write_image(snapshot_fd, resume_fd);
-			if (!error) {
-				splash.progress(100);
-#ifdef CONFIG_BOTH
-				if (!s2ram) {
-					suspend_shutdown(snapshot_fd);
-				} else {
-					/* If we die (and allow system to continue) between
-                                         * now and reset_signature(), very bad things will
-                                         * happen. */
-					error = suspend_to_ram(snapshot_fd);
-					if (error)
-						suspend_shutdown(snapshot_fd);
-					reset_signature(resume_fd);
-					free_swap_pages(snapshot_fd);
-					free_snapshot(snapshot_fd);
-					s2ram_resume();
-					goto Unfreeze;
-				}
-#else
-				suspend_shutdown(snapshot_fd);
-#endif
-			} else {
-				free_swap_pages(snapshot_fd);
-				free_snapshot(snapshot_fd);
-				image_size = 0;
-				error = -error;
-				if (error != ENOSPC)
-					break;
-			}
-		} else {
+		if (atomic_snapshot(snapshot_fd, &in_suspend)) {
 			error = errno;
 			break;
 		}
+		if (!in_suspend) {
+			/* first unblank the console, see console_codes(4) */
+			printf("\e[13]");
+			printf("%s: returned to userspace\n", my_name);
+			free_snapshot(snapshot_fd);
+			break;
+		}
+
+		error = write_image(snapshot_fd, resume_fd);
+		if (error) {
+			free_swap_pages(snapshot_fd);
+			free_snapshot(snapshot_fd);
+			image_size = 0;
+			error = -error;
+			if (error != ENOSPC)
+				break;
+		} else {
+			splash.progress(100);
+#ifdef CONFIG_BOTH
+			if (s2ram) {
+				/* If we die (and allow system to continue)
+				 * between now and reset_signature(), very bad
+				 * things will happen. */
+				error = suspend_to_ram(snapshot_fd);
+				if (error)
+					goto Shutdown;
+				reset_signature(resume_fd);
+				free_swap_pages(snapshot_fd);
+				free_snapshot(snapshot_fd);
+				s2ram_resume();
+				goto Unfreeze;
+			}
+Shutdown:
+#endif
+			suspend_shutdown(snapshot_fd);
+		}
 	} while (--attempts);
 
-	/* we only get here when we failed to suspend.
-	   Remember, suspend_shutdown() never returns! */
-	if (use_platform_suspend)
+	/* We get here during the resume or when we failed to suspend.
+	 * Remember, suspend_shutdown() never returns!
+	 */
+	if (shutdown_method == SHUTDOWN_METHOD_PLATFORM)
 		platform_finish(snapshot_fd);
 
 Unfreeze:
@@ -907,13 +985,13 @@ static int prepare_console(int *orig_vc, int *new_vc)
 		return fd;
 	error = ioctl(fd, VT_ACTIVATE, vt);
 	if (error) {
-		fprintf(stderr, "Could not activate the VT %d\n", vt);
+		suspend_error("Could not activate the VT %d.", vt);
 		fflush(stderr);
 		goto Close_fd;
 	}
 	error = ioctl(fd, VT_WAITACTIVE, vt);
 	if (error) {
-		fprintf(stderr, "VT %d activation failed\n", vt);
+		suspend_error("VT %d activation failed.", vt);
 		fflush(stderr);
 		goto Close_fd;
 	}
@@ -933,8 +1011,8 @@ static int prepare_console(int *orig_vc, int *new_vc)
 		tiocl[0] = TIOCL_SETKMSGREDIRECT;
 		tiocl[1] = vt;
 		if (ioctl(fd, TIOCLINUX, tiocl)) {
-			fprintf(stderr, "Failed to redirect kernel messages to VT %d\n"
-					"Reason: %s\n", vt, strerror(errno));
+			suspend_error("Failed to redirect kernel messages "
+				"to VT %d.", vt);
 			fflush(stderr);
 			set_kmsg_redirect = 0;
 		}
@@ -957,13 +1035,13 @@ static void restore_console(int fd, int orig_vc)
 
 	error = ioctl(fd, VT_ACTIVATE, orig_vc);
 	if (error) {
-		fprintf(stderr, "Could not activate the VT %d\n", orig_vc);
+		suspend_error("Could not activate the VT %d.", orig_vc);
 		fflush(stderr);
 		goto Close_fd;
 	}
 	error = ioctl(fd, VT_WAITACTIVE, orig_vc);
 	if (error) {
-		fprintf(stderr, "VT %d activation failed\n", orig_vc);
+		suspend_error("VT %d activation failed.", orig_vc);
 		fflush(stderr);
 	}
 	if (set_kmsg_redirect) {
@@ -977,39 +1055,6 @@ Close_fd:
 	close(fd);
 }
 
-static FILE *printk_file;
-
-static inline void open_printk(void)
-{
-	printk_file = fopen("/proc/sys/kernel/printk", "r+");
-}
-
-static inline int get_kernel_console_loglevel(void)
-{
-	int level = -1;
-
-	if (printk_file) {
-		rewind(printk_file);
-		fscanf(printk_file, "%d", &level);
-	}
-	return level;
-}
-
-static inline void set_kernel_console_loglevel(int level)
-{
-	if (printk_file) {
-		rewind(printk_file);
-		fprintf(printk_file, "%d\n", level);
-		fflush(printk_file);
-	}
-}
-
-static inline void close_printk(void)
-{
-	if (printk_file)
-		fclose(printk_file);
-}
-
 static FILE *swappiness_file;
 
 static inline void open_swappiness(void)
@@ -1053,6 +1098,7 @@ static void generate_key(void)
 	int ret, fd, rnd_fd;
 	struct RSA_data *rsa;
 	unsigned char *buf;
+	unsigned short size;
 	int j;
 
 	if (!key_data)
@@ -1091,50 +1137,56 @@ static void generate_key(void)
 		ret = gcry_ac_key_init(&rsa_pub, rsa_hd, GCRY_AC_KEY_PUBLIC,
 					rsa_data_set);
 
+	if (ret)
+		goto Destroy_data_set;
+
+	ret = gcry_ac_data_new(&key_set);
+	if (ret)
+		goto Destroy_key;
+
+	rnd_fd = open("/dev/urandom", O_RDONLY);
+	if (rnd_fd <= 0)
+		goto Destroy_key_set;
+
+	size = KEY_SIZE + CIPHER_BLOCK;
+	if (read(rnd_fd, key_data->key, size) != size)
+		goto Close_urandom;
+
+	gcry_mpi_scan(&mpi, GCRYMPI_FMT_USG, key_data->key, size, NULL);
+	ret = gcry_ac_data_encrypt(rsa_hd, 0, rsa_pub, mpi, &key_set);
+	gcry_mpi_release(mpi);
 	if (!ret) {
-		unsigned short size;
+		struct encrypted_key *key = &key_data->encrypted_key;
+		char *str;
+		size_t s;
 
-		ret = gcry_ac_data_new(&key_set);
-		if (ret)
-			goto Destroy_key;
-
-		rnd_fd = open("/dev/urandom", O_RDONLY);
-		if (rnd_fd <= 0)
-			goto Destroy_set;
-
-		size = KEY_SIZE + CIPHER_BLOCK;
-		if (read(rnd_fd, key_data->key, size) != size)
-			goto Close_urandom;
-
-		gcry_mpi_scan(&mpi, GCRYMPI_FMT_USG, key_data->key, size, NULL);
-		ret = gcry_ac_data_encrypt(rsa_hd, 0, rsa_pub, mpi, &key_set);
+		gcry_ac_data_get_index(key_set, GCRY_AC_FLAG_COPY, 0,
+						(const char **)&str, &mpi);
+		gcry_free(str);
+		ret = gcry_mpi_print(GCRYMPI_FMT_USG, key->data, KEY_DATA_SIZE,
+					&s, mpi);
 		gcry_mpi_release(mpi);
 		if (!ret) {
-			struct encrypted_key *key = &key_data->encrypted_key;
-			char *str;
-			size_t s;
-
-			gcry_ac_data_get_index(key_set, GCRY_AC_FLAG_COPY, 0,
-						(const char **)&str, &mpi);
-			gcry_free(str);
-			gcry_mpi_print(GCRYMPI_FMT_USG, key->data,
-					KEY_DATA_SIZE, &s, mpi);
-			gcry_mpi_release(mpi);
 			key->size = s;
 			use_RSA = 'y';
 		}
-Close_urandom:
-		close(rnd_fd);
-Destroy_set:
-		gcry_ac_data_destroy(key_set);
-Destroy_key:
-		gcry_ac_key_destroy(rsa_pub);
-	} else {
-		gcry_ac_data_destroy(rsa_data_set);
 	}
 
+Close_urandom:
+	close(rnd_fd);
+
+Destroy_key_set:
+	gcry_ac_data_destroy(key_set);
+
+Destroy_key:
+	gcry_ac_key_destroy(rsa_pub);
+
+Destroy_data_set:
+	gcry_ac_data_destroy(rsa_data_set);
+
 Free_rsa:
 	gcry_ac_close(rsa_hd);
+
 Close:
 	close(fd);
 }
@@ -1197,54 +1249,124 @@ static int lock_vt(void)
 static inline int get_config(int argc, char *argv[])
 {
 	static struct option options[] = {
-		{ "help",		no_argument,		NULL, 'h'},
-		{ "config",		required_argument,	NULL, 'f'},
-		{ "image_size",		required_argument,	NULL, 's'},
-		{ "resume_offset",	required_argument,	NULL, 'o'},
-		{ NULL,			0,			NULL,  0 }
+		   {
+		       "help\0\t\t\tthis text",
+		       no_argument,		NULL, 'h'
+		   },
+		   {
+		       "version\0\t\t\tversion information",
+		       no_argument,		NULL, 'V'
+		   },
+		   {
+		       "config\0\t\talternative configuration file.",
+		       required_argument,	NULL, 'f'
+		   },
+		   {
+		       "resume_device\0device that contains swap area",	
+		       required_argument,	NULL, 'r'
+		   },
+		   {
+		       "resume_offset\0offset of swap file in resume device.",	
+		       required_argument,	NULL, 'o'
+		   },
+		   {
+		       "image_size\0\tdesired size of the image.",
+		       required_argument,	NULL, 's'
+		   },
+		   {
+		       "parameter\0\toverride config file parameter.",
+		       required_argument,	NULL, 'P'
+		   },
+#ifdef CONFIG_BOTH
+		   HACKS_LONG_OPTS
+#endif
+		   { NULL,		0,			NULL,  0 }
 	};
 	int i, error;
 	char *conf_name = CONFIG_FILE;
-	int set_off = 0;
-	unsigned long long int off = 0;
-	int set_size = 0;
-	unsigned long int im_size = 0;
+	const char *optstring = "hVf:s:o:r:P:";
+	struct stat stat_buf;
+	int fail_missing_config = 0;
 
-	while ((i = getopt_long(argc, argv, "hf:s:o:", options, NULL)) != -1) {
+	/* parse only config file argument */
+	while ((i = getopt_long(argc, argv, optstring, options, NULL)) != -1) {
 		switch (i) {
 		case 'h':
-			usage("suspend", options);
-			exit(0);
+			usage(my_name, options, optstring);
+			exit(EXIT_SUCCESS);
+		case 'V':
+			version(my_name, NULL);
+			exit(EXIT_SUCCESS);
 		case 'f':
 			conf_name = optarg;
+			fail_missing_config = 1;
 			break;
-		case 's':
-			im_size = atoll(optarg);
-			set_size = 1;
-			break;
-		case 'o':
-			off = atoll(optarg);
-			set_off = 1;
-			break;
-		default:
-			usage("suspend", options);
-			return -EINVAL;
 		}
 	}
-	error = parse("suspend", conf_name, PARAM_NO, parameters);
-	if (error) {
-		fprintf(stderr, "suspend: Could not parse config file\n");
-		return error;
-	}
-	if (set_off)
-		resume_offset = off;
 
-	if (set_size)
-		pref_image_size = im_size;
+	if (stat(conf_name, &stat_buf)) {
+		if (fail_missing_config) {
+			fprintf(stderr, "%s: Could not stat configuration file\n",
+				my_name);
+			return -ENOENT;
+		}
+	}
+	else {
+		error = parse(my_name, conf_name, parameters);
+		if (error) {
+			fprintf(stderr, "%s: Could not parse config file\n", my_name);
+			return error;
+		}
+	}
+
+	optind = 0;
+	while ((i = getopt_long(argc, argv, optstring, options, NULL)) != -1) {
+		switch (i) {
+		case 'f':
+			/* already handled */
+			break;
+		case 's':
+			pref_image_size = atoll(optarg);
+			break;
+		case 'o':
+			resume_offset = atoll(optarg);
+			break;
+		case 'r':
+			strncpy(resume_dev_name, optarg, MAX_STR_LEN -1);
+			break;
+		case 'P':
+			error = parse_line(optarg, parameters);
+			if (error) {
+				fprintf(stderr, "%s: Could not parse config string '%s'\n", my_name, optarg);
+				return error;
+			}
+			break;
+		default:
+#ifdef CONFIG_BOTH
+			s2ram_add_flag(i, optarg);
+			break;
+#else
+			usage(my_name, options, optstring);
+			return -EINVAL;
+#endif
+		}
+	}
 
 	if (optind < argc)
 		strncpy(resume_dev_name, argv[optind], MAX_STR_LEN - 1);
 
+#ifdef CONFIG_BOTH
+	s2ram = s2ram_is_supported();
+	/* s2ram_is_supported returns EINVAL if there was something wrong
+	 * with the options that where added with s2ram_add_flag.
+	 * On any other error (unsupported) we will just continue with s2disk.
+	 */
+	if (s2ram == EINVAL)
+		return -EINVAL;
+	
+	s2ram = !s2ram;
+#endif
+
 	return 0;
 }
 
@@ -1258,6 +1380,8 @@ int main(int argc, char *argv[])
 	struct rlimit rlim;
 	static char chroot_path[MAX_STR_LEN];
 
+	my_name = basename(argv[0]);
+
 	/* Make sure the 0, 1, 2 descriptors are open before opening the
 	 * snapshot and resume devices
 	 */
@@ -1265,7 +1389,7 @@ int main(int argc, char *argv[])
 		ret = open("/dev/null", O_RDWR);
 		if (ret < 0) {
 			ret = errno;
-			fprintf(stderr, "suspend: Could not open /dev/null\n");
+			suspend_error("Could not open /dev/null.");
 			return ret;
 		}
 	} while (ret < 3);
@@ -1278,49 +1402,68 @@ int main(int argc, char *argv[])
 	if (compute_checksum != 'y' && compute_checksum != 'Y')
 		compute_checksum = 0;
 #ifdef CONFIG_COMPRESS
-	if (compress != 'y' && compress != 'Y')
-		compress = 0;
+	if (do_compress != 'y' && do_compress != 'Y') {
+		do_compress = 0;
+	} else if (lzo_init() != LZO_E_OK) {
+		suspend_error("Failed to initialize LZO. "
+				"Compression disabled.\n");
+		do_compress = 0;
+	}
 #endif
 #ifdef CONFIG_ENCRYPT
-	if (encrypt != 'y' && encrypt != 'Y')
-		encrypt = 0;
+	if (do_encrypt != 'y' && do_encrypt != 'Y')
+		do_encrypt = 0;
 #endif
 	if (splash_param != 'y' && splash_param != 'Y')
 		splash_param = 0;
+	else
+		splash_param = SPL_SUSPEND;
 
 	if (early_writeout != 'n' && early_writeout != 'N')
 		early_writeout = 1;
 
-	use_platform_suspend = !strcmp(shutdown_method, "platform");
+	if (!strcmp (shutdown_method_value, "shutdown")) {
+		shutdown_method = SHUTDOWN_METHOD_SHUTDOWN;
+	}
+	else if (!strcmp (shutdown_method_value, "platform")) {
+		shutdown_method = SHUTDOWN_METHOD_PLATFORM;
+	}
+	else if (!strcmp (shutdown_method_value, "reboot")) {
+		shutdown_method = SHUTDOWN_METHOD_REBOOT;
+	}
 
 	page_size = getpagesize();
 	buffer_size = BUFFER_PAGES * page_size;
 
 	mem_size = 3 * page_size + buffer_size;
-#ifdef CONFIG_ENCRYPT
-	if (encrypt)
-		mem_size += buffer_size;
+#ifdef CONFIG_COMPRESS
+	if (do_compress)
+		mem_size += LZO1X_1_MEM_COMPRESS;
 #endif
-	mem_pool = malloc(mem_size);
-	if (!mem_pool) {
-		ret = errno;
-		fprintf(stderr, "suspend: Could not allocate memory\n");
-		return ret;
-	}
 #ifdef CONFIG_ENCRYPT
-	if (encrypt) {
-		printf("suspend: libgcrypt version: %s\n",
+	if (do_encrypt) {
+		printf("%s: libgcrypt version: %s\n", my_name,
 			gcry_check_version(NULL));
 		gcry_control(GCRYCTL_INIT_SECMEM, page_size, 0);
 		ret = gcry_cipher_open(&cipher_handle, IMAGE_CIPHER,
 				GCRY_CIPHER_MODE_CFB, GCRY_CIPHER_SECURE);
 		if (ret) {
-			fprintf(stderr, "suspend: libgcrypt error %s\n",
+			fprintf(stderr, "%s: libgcrypt error %s\n", my_name, 
 				gcry_strerror(ret));
-			encrypt = 0;
+			do_encrypt = 0;
+		} else {
+			mem_size += buffer_size;
 		}
 	}
-	if (encrypt) {
+#endif
+	mem_pool = malloc(mem_size);
+	if (!mem_pool) {
+		ret = errno;
+		suspend_error("Could not allocate memory.");
+		return ret;
+	}
+#ifdef CONFIG_ENCRYPT
+	if (do_encrypt) {
 		mem_size -= buffer_size;
 		key_data = (struct key_data *)((char *)mem_pool + mem_size);
 		generate_key();
@@ -1332,77 +1475,80 @@ int main(int argc, char *argv[])
 
 	if (mlockall(MCL_CURRENT | MCL_FUTURE)) {
 		ret = errno;
-		fprintf(stderr, "suspend: Could not lock myself\n");
+		suspend_error("Could not lock myself.");
 		return ret;
 	}
 
 	snprintf(chroot_path, MAX_STR_LEN, "/proc/%d", getpid());
 	if (mount("none", chroot_path, "tmpfs", 0, NULL)) {
 		ret = errno;
-		fprintf(stderr, "suspend: Could not mount tmpfs on %s\n", chroot_path);
+		suspend_error("Could not mount tmpfs on %s.", chroot_path);
 		return ret;
 	}
 
 	ret = 0;
 	if (stat(resume_dev_name, &stat_buf)) {
-		fprintf(stderr, "suspend: Could not stat the resume device file\n");
+		suspend_error("Could not stat the resume device file.");
 		ret = ENODEV;
 		goto Umount;
 	}
 	if (!S_ISBLK(stat_buf.st_mode)) {
-		fprintf(stderr, "suspend: Invalid resume device\n");
+		fprintf(stderr, "%s: Invalid resume device\n", my_name);
 		ret = EINVAL;
 		goto Umount;
 	}
 	if (chdir(chroot_path)) {
 		ret = errno;
-		fprintf(stderr, "suspend: Could not change directory to %s\n",
+		suspend_error("Could not change directory to %s.",
 			chroot_path);
 		goto Umount;
 	}
 	resume_dev = stat_buf.st_rdev;
 	if (mknod("resume", S_IFBLK | 0600, resume_dev)) {
 		ret = errno;
-		fprintf(stderr, "suspend: Could not create %s/%s\n",
-			chroot_path, "resume");
+		suspend_error("Could not create %s/%s.", chroot_path, "resume");
 		goto Umount;
 	}
 	resume_fd = open("resume", O_RDWR);
 	if (resume_fd < 0) {
 		ret = errno;
-		fprintf(stderr, "suspend: Could not open the resume device\n");
+		suspend_error("Could not open the resume device.");
 		goto Umount;
 	}
 
 	if (stat(snapshot_dev_name, &stat_buf)) {
-		fprintf(stderr, "suspend: Could not stat the snapshot device file\n");
+		suspend_error("Could not stat the snapshot device file.");
 		ret = ENODEV;
 		goto Close_resume_fd;
 	}
 
 	if (!S_ISCHR(stat_buf.st_mode)) {
-		fprintf(stderr, "suspend: Invalid snapshot device\n");
+		fprintf(stderr, "%s: Invalid snapshot device\n", my_name);
 		ret = EINVAL;
 		goto Close_resume_fd;
 	}
 	snapshot_fd = open(snapshot_dev_name, O_RDONLY);
 	if (snapshot_fd < 0) {
 		ret = errno;
-		fprintf(stderr, "suspend: Could not open the snapshot device\n");
+		suspend_error("Could not open the snapshot device.");
 		goto Close_resume_fd;
 	}
 
 	if (set_swap_file(snapshot_fd, resume_dev, resume_offset)) {
 		ret = errno;
-		fprintf(stderr, "suspend: Could not use the resume device "
-			"(try swapon -a)\n");
+		suspend_error("Could not use the resume device "
+			"(try swapon -a).");
 		goto Close_snapshot_fd;
 	}
 
 	vt_fd = prepare_console(&orig_vc, &suspend_vc);
 	if (vt_fd < 0) {
 		ret = errno;
-		fprintf(stderr, "suspend: Could not open a virtual terminal\n");
+		if (vt_fd == -ENOTTY)
+			suspend_error("No local tty. Remember to specify local " \
+					"console after the remote.");
+		else
+			suspend_error("Could not open a virtual terminal.");
 		goto Close_snapshot_fd;
 	}
 
@@ -1410,18 +1556,19 @@ int main(int argc, char *argv[])
 
 	if (lock_vt() < 0) {
 		ret = errno;
-		fprintf(stderr, "suspend: Could not lock the terminal\n");
+		suspend_error("Could not lock the terminal.");
 		goto Restore_console;
 	}
 
 	splash.progress(5);
 
 #ifdef CONFIG_BOTH
-	/* If s2ram_prepare returns != 0, better not try to suspend to RAM */
-	s2ram = !s2ram_prepare();
+	/* If s2ram_hacks returns != 0, better not try to suspend to RAM */
+	if (s2ram) 
+		s2ram = !s2ram_hacks();
 #endif
 #ifdef CONFIG_ENCRYPT
-        if (encrypt && ! use_RSA) {
+        if (do_encrypt && ! use_RSA) {
                 splash.read_password((char *)mem_pool,1);
                 strncpy(password,(char *)mem_pool,PASS_SIZE);
         }
@@ -1467,12 +1614,12 @@ Close_resume_fd:
 Umount:
 	if (chdir("/")) {
 		ret = errno;
-		fprintf(stderr, "suspend: Could not change directory to /\n");
+		suspend_error("Could not change directory to /");
 	} else {
 		umount(chroot_path);
 	}
 #ifdef CONFIG_ENCRYPT
-	if (encrypt)
+	if (do_encrypt)
 		gcry_cipher_close(cipher_handle);
 #endif
 	free(mem_pool);
